/*
 *  Copyright 2008 CNRS
 *
 *  Author: Florent Lamiraux
 */

#include <iostream>

#include "kppInterfaceDisplay/kppInterfaceDisplay.h" 

#include "KineoGUI/kppMainWindowUICommandFactory.h"
#include "KineoController/kppUICommandList.h"
#include "KineoController/kppUICommand.h"

#include "kppInterface/kppCommandStartCorbaServer.h"
#include "kppInterfaceWalk/kppCommandLoadHRP2.h"
#include "kppInterfaceWalk/openHrpConfig.h"

#include "KineoWorks2/kwsSMLinear.h"
#include "kppInterfaceDisplay/kppCommandCreateGenoMTraj.h"
#include "kppInterfaceDisplay/kppCommandUpdateGenoMTraj.h"

#include "posterLib.h"
#include "trajPoster.h"

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cerr << "kppInterfaceDisplay:" << x << std::endl
#define ODEBUG1(x) std::cerr << "kppInterfaceDisplay:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "kppInterfaceDisplay:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

CkppInterfaceDisplayShPtr CkppInterfaceDisplay::create()
{
  std::string pathToHrp2(KPPINTERFACEWALK_OPENHRP_PREFIX);
  pathToHrp2 += "/Controller/IOserver/robot/HRP2JRL/model/HRP2JRLmain.wrl";
  // Create an ChppWalkPlanner object.
  ODEBUG2(" path to HRP2: " << pathToHrp2);
  // Create an ChppWalkPlanner object.
  ChppWalkPlanner *planner = new ChppWalkPlanner(0.2, 0.2, 0.2, 0.05,
						 pathToHrp2.c_str());
  CkppInterfaceDisplayShPtr shPtr;

  if (!planner) {
    ODEBUG1(":createHppWalkPlanner: allocation of ChppWalkPlanner object failed.");
    return shPtr;
  }

  CkppInterfaceDisplay* ptr = new CkppInterfaceDisplay(planner);
  return CkppInterfaceDisplayShPtr(ptr);
}

CkppInterfaceDisplay::CkppInterfaceDisplay(ChppWalkPlanner* inHppPlanner) :
  CkppInterfaceWalk(inHppPlanner)
{
}

CkppInterfaceDisplay::~CkppInterfaceDisplay()
{
}

// ==========================================================================

void CkppInterfaceDisplay::getMenuUICommandLists(const CkppMainWindowUICommandFactoryConstShPtr& inCommandFactory,
						 std::vector< CkppUICommandListShPtr > & outMenuCommandListVector)
{
  /*
    Get and store pointer to the main window controller. This is the only place where we
    have access to this object
  */
  getMainWindowController(inCommandFactory);
  // ---  LIST --- //
  CkppUICommandListShPtr hppUICommandList = CkppUICommandList::create("Hpp/Display");
 

  // --- COMMAND --- //
 
  attCommandLoadHRP2 = CkppUICommand::create(CkppCommandLoadHRP2::create(this),
						  inCommandFactory->environment(),
						  "Load HRP2",
						  "Load HRP2");

  attStartCorbaServerCommand = CkppUICommand::create(CkppCommandStartCorbaServer::create(this),
						     inCommandFactory->environment(),
						     "Start CORBA Server",
						     "start corba server");

  attCommandCreateGenoMTraj = CkppUICommand::create(CkppCommandCreateGenoMTraj::create(this),
						    inCommandFactory->environment(),
						    "Create Trajectory",
						    "Create trajectory associated to a GenoM poster");

  attCommandUpdateGenoMTraj = CkppUICommand::create(CkppCommandUpdateGenoMTraj::create(this),
						    inCommandFactory->environment(),
						    "Update Trajectory",
						    "Update trajectory by reading GenoM poster");

  
  hppUICommandList->appendCommand(attCommandLoadHRP2) ; 
  hppUICommandList->appendCommand(attStartCorbaServerCommand) ; 
  hppUICommandList->appendCommand(attCommandCreateGenoMTraj);
  hppUICommandList->appendCommand(attCommandUpdateGenoMTraj);

  outMenuCommandListVector.push_back(hppUICommandList);
 
}

void CkppInterfaceDisplay::addPosterNameToMapping(const CkwsPathShPtr& inPath, std::string inPosterName)
{
  if (attPosterNameMap.count(inPath) != 0) {
    ODEBUG2(":addPosterNameToMapping: path already linked to a poster.");
  }
  attPosterNameMap[inPath] = inPosterName;
}

ktStatus CkppInterfaceDisplay::getPosterName(CkwsPathShPtr inPath, std::string& outPosterName)
{
  if (attPosterNameMap.count(inPath) != 1) {
    ODEBUG1(":getPosterName: no poster linked to this path.");
    return KD_ERROR;
  }
  outPosterName = attPosterNameMap[inPath];
  return KD_OK;
}

ktStatus CkppInterfaceDisplay::updatePathFromPoster(const CkwsPathShPtr& inPath)
{
  CkwsDeviceShPtr robot = inPath->device();

  // Find poster name associated to path
  std::string trajPosterName;
  if (getPosterName(inPath, trajPosterName) != KD_OK) {
    return KD_ERROR;
  }

  // Get poster id associated to poster name.
  POSTER_ID posterId;
  if (posterFind (const_cast<char*>(trajPosterName.c_str()), 
		  &posterId) != OK) {
    ODEBUG1(":updatePathFromPoster: poster " << trajPosterName << " not found.");
    return KD_ERROR;
  }

  // Read poster
  ROBOT_TRAJ_POSTER_STR robotTraj;
  if (posterRead(posterId, 0, &robotTraj, sizeof(ROBOT_TRAJ_POSTER_STR)) !=
      sizeof(ROBOT_TRAJ_POSTER_STR)) {
    ODEBUG1(":updatePathFromPoster: failed to read " << trajPosterName << ".");
    return KD_ERROR;
  }

  // Test that data in poster is consistant
  if (robotTraj.nbDof <= 0 || robotTraj.nbConfig <= 0 || 
      robotTraj.nbConfig*robotTraj.nbDof > ROBOT_TRAJ_MAX_NB_DOF) {
    ODEBUG1(":updatePathFromPoster: size of trajectory is not valid.");
    ODEBUG1(":nbDof = " << robotTraj.nbDof);
    ODEBUG1(":robotTraj.nbConfig = " << robotTraj.nbConfig);
    return KD_ERROR;
  }

  ODEBUG2(":updatePathFromPoster: robotTraj.nbDof = " << robotTraj.nbDof);
  ODEBUG2(":updatePathFromPoster: robotTraj.nbConfig = " << robotTraj.nbConfig);

  // Test that dimension of robot is consistant
  if ((unsigned int)robotTraj.nbDof != robot->countDofs()) {
    ODEBUG1(":updatePathFromPoster: nb dofs in poster and in robot do not match.");
    return KD_ERROR;
  }

  // Clear previous path
  inPath->clear();

  // Create a linear steering method 
  CkwsSMLinearShPtr steeringMethod = CkwsSMLinear::create();

  // Build first configuration
  std::vector<double> dofVector;
  for (unsigned int iDof=0; iDof < (unsigned int)robotTraj.nbDof; iDof++) {
    dofVector.push_back(robotTraj.dof[iDof]);
  }
  CkwsConfig initConfig(robot, dofVector);
  CkwsConfig endConfig(robot);

  // Loop over sample configurations
  for (unsigned int iConfig=1; iConfig < (unsigned int) robotTraj.nbConfig; iConfig++) {
    // Build end configuration of direct path
    dofVector.clear();
    for (unsigned int iDof=0; iDof < (unsigned int)robotTraj.nbDof; iDof++) {
      dofVector.push_back(robotTraj.dof[iConfig*robotTraj.nbDof + iDof]);
    }
    endConfig.setDofValues(dofVector);
    CkwsDirectPathShPtr directPath = steeringMethod->makeDirectPath(initConfig, endConfig);
    if (!directPath){
      ODEBUG1(":updatePathFromPoster: direct path does not exist");
      ODEBUG1(":updatePathFromPoster: iConfig = " << iConfig << ".");
      return KD_ERROR;
    }
    if (inPath->appendDirectPath(directPath) != KD_OK) {
      ODEBUG1(":updatePathFromPoster: could not append direct path.");
      ODEBUG1(":updatePathFromPoster: iConfig = " << iConfig << ".");
      return KD_ERROR;
    }
    initConfig.setDofValues(dofVector);
  }

  return KD_OK;
}


// ------------- initializing module ------------------
// extern Function : First called by KineoWorks to Load the Module

int initializeModule(CkppModuleInterfaceShPtr& o_moduleInterface)
{
  cerr<<" initializing module ... "<<endl;
  o_moduleInterface = CkppInterfaceDisplay::create();
  return 0;
}

